@@ -1,5 +1,5 @@
 /*
-** Copyright 2007-2013 Double Precision, Inc.
+** Copyright 2007-2018 Double Precision, Inc.
 ** See COPYING for distribution information.
 */
 #include	"config.h"
@@ -13,7 +13,6 @@
 #include	<gnutls/extra.h>
 #endif
 #include	<gnutls/x509.h>
-#include	<gnutls/openpgp.h>
 #include	<stdio.h>
 #include	<string.h>
 #include	<stdlib.h>
@@ -127,6 +126,7 @@ struct ssl_context_t {
 	const char *priority_list;
 
 	char *certfile;
+	char *keyfile;
 	char *dhfile;
 
 	char *trustcerts;
@@ -146,9 +146,6 @@ struct ssl_handle_t {
 	gnutls_session_t session;
 
 	gnutls_x509_privkey_t x509_key;
-
-	gnutls_openpgp_crt_t pgp_crt;
-	gnutls_openpgp_privkey_t pgp_key;
 };
 
 static void nonsslerror(struct tls_info *info, const char *pfix)
@@ -271,6 +268,17 @@ ssl_context tls_create(int isserver, const struct tls_info *info)
 			free(certfile);
 	}
 
+	if ((certfile=strdup(safe_getenv(p, "TLS_PRIVATE_KEYFILE", ""))) != NULL &&
+	    *certfile)
+	{
+		p->keyfile=certfile;
+	}
+	else
+	{
+		if (certfile)
+			free(certfile);
+	}
+
 	switch (*safe_getenv(p, "TLS_VERIFYPEER", "P")) {
 	case 'n':
 	case 'N':
@@ -324,6 +332,8 @@ void tls_destroy(ssl_context p)
 {
 	if (p->certfile)
 		free(p->certfile);
+	if (p->keyfile)
+		free(p->keyfile);
 	if (p->dhfile)
 		free(p->dhfile);
 	if (p->trustcerts)
@@ -367,7 +377,7 @@ static int read_cert_dir(const char *cert_dir,
 
 		strcat(strcat(strcpy(buf, cert_dir), "/"), de->d_name);
 
-		if (lstat(buf, &stat_buf) < 0 || !S_ISREG(stat_buf.st_mode))
+		if (stat(buf, &stat_buf) < 0 || !S_ISREG(stat_buf.st_mode))
 		{
 			free(buf);
 			continue;
@@ -475,16 +485,7 @@ static void tls_free_session_keys(ssl_handle ssl)
 	if (ssl->x509_key)
 		gnutls_x509_privkey_deinit(ssl->x509_key);
 
-	if (ssl->pgp_crt)
-		gnutls_openpgp_key_deinit(ssl->pgp_crt);
-
-	if (ssl->pgp_key)
-		gnutls_openpgp_privkey_deinit(ssl->pgp_key);
-
 	ssl->x509_key=NULL;
-	ssl->pgp_crt=NULL;
-	ssl->pgp_key=NULL;
-
 }
 
 static void tls_free_session(ssl_handle ssl)
@@ -537,7 +538,15 @@ static int chk_error(int rc, ssl_handle ssl, int fd, fd_set *r, fd_set *w,
 
 	if (result_rc)
 	{
-		(*ssl->info_cpy.tls_err_msg)(gnutls_strerror(rc),
+		const char *errmsg=gnutls_strerror(rc);
+
+#ifdef GNUTLS_E_PREMATURE_TERMINATION
+		if (rc == GNUTLS_E_PREMATURE_TERMINATION)
+		{
+			errmsg="DEBUG: Unexpected SSL connection shutdown.";
+		}
+#endif
+		(*ssl->info_cpy.tls_err_msg)(errmsg,
 					     ssl->info_cpy.app_data);
 		*result_rc= -1;
 	}
@@ -672,94 +681,6 @@ static int verify_client(ssl_handle ssl, int fd)
 
 		gnutls_x509_crt_deinit (cert);
 	}
-	else if (gnutls_certificate_type_get(ssl->session)==GNUTLS_CRT_OPENPGP)
-	{
-		gnutls_openpgp_crt_t cert;
-
-		if (gnutls_openpgp_key_init(&cert) < 0)
-		{
-			(*ssl->info_cpy.tls_err_msg)
-				("Error initializing certificate",
-				 ssl->info_cpy.app_data);
-			gnutls_openpgp_key_deinit(cert);
-			return GNUTLS_E_INSUFFICIENT_CREDENTIALS;
-		}
-
-		if (gnutls_openpgp_key_import(cert, &cert_list[0],
-					      GNUTLS_OPENPGP_FMT_RAW) < 0)
-		{
-			(*ssl->info_cpy.tls_err_msg)
-				("Error parsing certificate",
-				 ssl->info_cpy.app_data);
-			gnutls_openpgp_key_deinit (cert);
-			return GNUTLS_E_INSUFFICIENT_CREDENTIALS;
-		}
-
-		if (gnutls_openpgp_key_get_creation_time(cert) > time(NULL))
-		{
-			(*ssl->info_cpy.tls_err_msg)
-				("Certificate not activated",
-				 ssl->info_cpy.app_data);
-			gnutls_openpgp_key_deinit (cert);
-			return GNUTLS_E_INSUFFICIENT_CREDENTIALS;
-		}
-
-		if (gnutls_openpgp_key_get_expiration_time(cert) < time(NULL))
-		{
-			(*ssl->info_cpy.tls_err_msg)
-				("Expired certificate",
-				 ssl->info_cpy.app_data);
-			gnutls_openpgp_key_deinit (cert);
-			return GNUTLS_E_INSUFFICIENT_CREDENTIALS;
-		}
-
-		if (ssl->info_cpy.peer_verify_domain &&
-		    *ssl->info_cpy.peer_verify_domain &&
-		    !gnutls_openpgp_key_check_hostname(cert,
-						       ssl->info_cpy
-						       .peer_verify_domain))
-						      
-		{
-			char *hostname;
-			size_t hostnamesiz=0;
-			const char *errmsg_txt=
-				"Certificate owner mismatch: ";
-			char *errmsg_buf;
-
-			gnutls_openpgp_key_get_name(cert, 0, NULL,
-						    &hostnamesiz);
-
-			hostname=malloc(hostnamesiz);
-
-			if (hostname)
-			{
-				*hostname=0;
-				gnutls_openpgp_key_get_name(cert,
-							    0, hostname,
-							    &hostnamesiz);
-			}
-
-			errmsg_buf=malloc(strlen(errmsg_txt)+
-					  strlen(hostname ?
-						 hostname:"")+100);
-
-			if (errmsg_buf)
-				strcat(strcpy(errmsg_buf, errmsg_txt),
-					       hostname ?
-					       hostname:"(unknown)");
-
-			(*ssl->info_cpy.tls_err_msg)
-				(errmsg_buf ? errmsg_buf:strerror(errno),
-				 ssl->info_cpy.app_data);
-			if (errmsg_buf)
-				free(errmsg_buf);
-			if (hostname)
-				free(hostname);
-			gnutls_openpgp_key_deinit (cert);
-			return GNUTLS_E_INSUFFICIENT_CREDENTIALS;
-		}
-		gnutls_openpgp_key_deinit (cert);
-	}
 	else
 	{
 		(*ssl->info_cpy.tls_err_msg)
@@ -784,7 +705,7 @@ static int dohandshake(ssl_handle ssl, int fd, fd_set *r, fd_set *w)
 	{
 		ssl->info_cpy.connect_interrupted=0;
 
-		
+
 		if (verify_client(ssl, fd))
 			return -1;
 
@@ -813,9 +734,6 @@ static char *check_cert(const char *filename,
 
 		strcat(strcat(strcpy(p, filename), "."), req_dn);
 
-		if (cert_type == GNUTLS_CRT_OPENPGP)
-			strcat(p, ".pgp");
-
 		if (access(p, R_OK) == 0)
 			return p;
 
@@ -837,9 +755,6 @@ static char *check_cert(const char *filename,
 
 		strcpy(p, filename);
 
-		if (cert_type == GNUTLS_CRT_OPENPGP)
-			strcat(p, ".pgp");
-
 		if (access(p, R_OK) == 0)
 			return p;
 
@@ -848,6 +763,14 @@ static char *check_cert(const char *filename,
 	return NULL;
 }
 
+static char *check_key(const char *filename,
+			gnutls_certificate_type_t cert_type,
+			const char *req_dn,
+			int isvirtual)
+{
+	return check_cert(filename, cert_type, req_dn, isvirtual);
+}
+
 static int read_file(const char *file,
 		     gnutls_datum_t *filebuf)
 {
@@ -889,10 +812,12 @@ static void release_file(gnutls_datum_t *filebuf)
 static int set_cert(ssl_handle ssl,
 		    gnutls_session_t session,
 		    gnutls_retr2_st *st,
-		    const char *certfilename)
+		    const char *certfilename,
+		    const char *keyfilename)
 {
 	int rc;
 	gnutls_datum_t filebuf;
+	gnutls_datum_t keyfilebuf;
 	unsigned int cert_cnt;
 
 	st->ncerts=0;
@@ -902,15 +827,34 @@ static int set_cert(ssl_handle ssl,
 	if ((rc=read_file(certfilename, &filebuf)) < 0)
 		return rc;
 
+	if (keyfilename)
+	{
+		if ((rc=read_file(keyfilename, &keyfilebuf)) < 0)
+		{
+			release_file(&filebuf);
+			return rc;
+		}
+	}
+
 	switch (st->cert_type) {
 	case GNUTLS_CRT_X509:
 
 		cert_cnt=0;
 
-		if ((rc=gnutls_x509_privkey_init(&ssl->x509_key)) < 0 ||
-		    (rc=gnutls_x509_privkey_import(ssl->x509_key, &filebuf,
-						   GNUTLS_X509_FMT_PEM)) < 0)
-			break;
+		if (keyfilename)
+		{
+			if ((rc=gnutls_x509_privkey_init(&ssl->x509_key)) < 0 ||
+			    (rc=gnutls_x509_privkey_import(ssl->x509_key, &keyfilebuf,
+							   GNUTLS_X509_FMT_PEM)) < 0)
+				break;
+		}
+		else
+		{
+			if ((rc=gnutls_x509_privkey_init(&ssl->x509_key)) < 0 ||
+			    (rc=gnutls_x509_privkey_import(ssl->x509_key, &filebuf,
+							   GNUTLS_X509_FMT_PEM)) < 0)
+				break;
+		}
 
 		rc=gnutls_x509_crt_list_import(NULL, &cert_cnt,
 					       &filebuf,
@@ -939,26 +883,14 @@ static int set_cert(ssl_handle ssl,
 		ssl->x509_key=0;
 		st->deinit_all=1;
 
-		break;
-	case GNUTLS_CRT_OPENPGP:
-		if ((rc=gnutls_openpgp_key_init(&ssl->pgp_crt)) < 0 ||
-		    (rc=gnutls_openpgp_privkey_init(&ssl->pgp_key)) < 0 ||
-		    (rc=gnutls_openpgp_key_import(ssl->pgp_crt, &filebuf,
-						  GNUTLS_OPENPGP_FMT_BASE64))
-		    < 0 ||
-		    (rc=gnutls_openpgp_privkey_import(ssl->pgp_key, &filebuf,
-						      GNUTLS_OPENPGP_FMT_BASE64,
-						      NULL, 0)) < 0)
-			break;
-		st->cert.pgp=ssl->pgp_crt;
-		st->ncerts=1;
-		st->key.pgp=ssl->pgp_key;
 		break;
 	default:
 		break;
 	}
 
 	release_file(&filebuf);
+	if (keyfilename)
+		release_file(&keyfilebuf);
 	return 0;
 }
 
@@ -975,6 +907,7 @@ static int get_server_cert(gnutls_session_t session,
 	size_t vhost_size;
 	unsigned int type=GNUTLS_NAME_DNS;
 	char *certfilename=NULL;
+	char *keyfilename=NULL;
 	int rc;
 
 	st->cert_type=gnutls_certificate_type_get(session);
@@ -1003,13 +936,18 @@ static int get_server_cert(gnutls_session_t session,
 
 		for (p=vhost_buf; *p; p++)
 			if (*p == '/')
-				*p='.';
+				*p='.'; /* Script kiddie check */
 
 		if (ssl->ctx->certfile)
 			certfilename=check_cert(ssl->ctx->certfile,
 						st->cert_type,
 						vhost_buf, 1);
 
+		if (ssl->ctx->keyfile)
+			keyfilename=check_key(ssl->ctx->keyfile,
+						st->cert_type,
+						vhost_buf, 1);
+
 		if (certfilename)
 			break;
 	}
@@ -1022,13 +960,25 @@ static int get_server_cert(gnutls_session_t session,
 						safe_getenv(ssl->ctx,
 							    "TCPLOCALIP", ""),
 						0);
+		if (ssl->ctx->keyfile)
+			keyfilename=check_key(ssl->ctx->keyfile,
+						st->cert_type,
+						safe_getenv(ssl->ctx,
+							    "TCPLOCALIP", ""),
+						0);
 	}
 
 	if (!certfilename)
+	{
+		if (ssl->ctx->certfile)
+			nonsslerror(&ssl->info_cpy, ssl->ctx->certfile);
 		return 0;
+	}
 
-	rc=set_cert(ssl, session, st, certfilename);
+	rc=set_cert(ssl, session, st, certfilename, keyfilename);
 	free(certfilename);
+	if (keyfilename)
+		free(keyfilename);
 	return rc;
 }
 
@@ -1181,6 +1131,7 @@ static int get_client_cert(gnutls_session_t session,
 	ssl_handle ssl=(ssl_handle)gnutls_session_get_ptr(session);
 	int rc;
 	char *certfilename=NULL;
+	char *keyfilename=NULL;
 
 	rc= 0;
 	st->cert_type=gnutls_certificate_type_get(session);
@@ -1189,12 +1140,16 @@ static int get_client_cert(gnutls_session_t session,
 		certfilename=check_cert(ssl->ctx->certfile,
 					st->cert_type, "", 0);
 
+	if (ssl->ctx->keyfile)
+		keyfilename=check_key(ssl->ctx->keyfile,
+					st->cert_type, "", 0);
+
 	st->ncerts=0;
 	st->deinit_all=0;
 
 	if (certfilename)
 	{
-		rc=set_cert(ssl, session, st, certfilename);
+		rc=set_cert(ssl, session, st, certfilename, keyfilename);
 		free(certfilename);
 	}
 	else
@@ -1204,6 +1159,10 @@ static int get_client_cert(gnutls_session_t session,
 		if (rc > 0)
 			rc=0;
 	}
+
+	if (keyfilename)
+		free(keyfilename);
+
 	return rc;
 }
 
@@ -1273,7 +1232,7 @@ static int do_cache_remove(void *rec, size_t recsize, int *doupdate, void *arg)
 	}
 	return 0;
 }
-	
+
 static int db_remove_func(void *dummy, gnutls_datum_t key)
 {
 	tls_cache_walk(((ssl_handle)dummy)->info_cpy.tlscache,
@@ -1444,9 +1403,9 @@ ssl_handle tls_connect(ssl_context ctx, int fd)
 	gnutls_session_set_ptr(ssl->session, ssl);
 
         gnutls_handshake_set_private_extensions(ssl->session, 1);
-        gnutls_certificate_set_verify_flags(ssl->xcred, 
+        gnutls_certificate_set_verify_flags(ssl->xcred,
                                             GNUTLS_VERIFY_ALLOW_X509_V1_CA_CRT |
-                                            
+
                                             /*
                                             GNUTLS_VERIFY_DO_NOT_ALLOW_SAME |
                                             GNUTLS_VERIFY_ALLOW_ANY_X509_V1_CA_C
@@ -1567,7 +1526,7 @@ int	tls_transfer(struct tls_transfer_info *t, ssl_handle ssl, int fd,
 			return 1;
 		}
 
-			
+
 		t->shutdown_interrupted=0;
 		t->shutdown= -1;
 		return -1;
@@ -1670,10 +1629,12 @@ int tls_connecting(ssl_handle ssl)
 }
 
 static const char *dump_dn(gnutls_x509_crt_t cert,
-			   int (*get_dn_func)(gnutls_x509_crt_t cert, int indx,
+			   int (*get_dn_func)(gnutls_x509_crt_t cert,
+					      unsigned indx,
 					      void *oid, size_t * sizeof_oid),
 			   int (*get_dnval_func)(gnutls_x509_crt_t cert,
-						 const char *oid, int indx,
+						 const char *oid,
+						 unsigned indx,
 						 unsigned int raw_flag,
 						 void *buf, size_t *sizeof_buf),
 			   void (*dump_func)(const char *, int cnt, void *),
@@ -1717,7 +1678,7 @@ static const char *dump_dn(gnutls_x509_crt_t cert,
 			free(oidname);
 			return gnutls_strerror(rc);
 		}
-	
+
 		vidx=0;
 
 		while (bufsiz=0,
@@ -1781,7 +1742,7 @@ static const char *dump_dn(gnutls_x509_crt_t cert,
 			++vidx;
 		}
 	}
-	
+
 	free(oidval);
 	free(oidname);
 	return NULL;
@@ -1867,10 +1828,8 @@ static void dump_cipher_name(gnutls_session_t session,
 		(*dump_func)(cipher_name, -1, dump_arg);
 	else
 	{
-		gnutls_compression_method_t comp;
-
 		(*dump_func)(gnutls_kx_get_name(kx_algo), -1, dump_arg);
-		
+
 		(*dump_func)("-", 1, dump_arg);
 		(*dump_func)(gnutls_certificate_type_get_name(gnutls_certificate_type_get(session)),
 			     -1, dump_arg);
@@ -1879,14 +1838,6 @@ static void dump_cipher_name(gnutls_session_t session,
 		(*dump_func)(gnutls_cipher_get_name(cipher_algo), -1,
 			     dump_arg);
 
-		if ((comp=gnutls_compression_get(session))
-		    != GNUTLS_COMP_NULL)
-		{
-			(*dump_func)("/", 1, dump_arg);
-			(*dump_func)(gnutls_compression_get_name(comp),
-				     -1, dump_arg);
-		}
-
 		(*dump_func)("-", 1, dump_arg);
 		(*dump_func)(gnutls_mac_get_name(gnutls_mac_get(session)),
 			     -1, dump_arg);
@@ -1923,6 +1874,8 @@ static void tls_dump_connection_info_x509(ssl_handle ssl,
 		{
 			time_t notbefore;
 			time_t notafter;
+			char buffer[256];
+			unsigned j;
 
 			(*dump_func)("Subject:\n", -1, dump_arg);
 
@@ -1932,6 +1885,29 @@ static void tls_dump_connection_info_x509(ssl_handle ssl,
 				dump_func, dump_arg);
 			(*dump_func)("\n", 1, dump_arg);
 
+			for (j=0; ; ++j)
+			{
+				size_t s=sizeof(buffer);
+				enum gnutls_x509_subject_alt_name_t t;
+
+				t=gnutls_x509_crt_get_subject_alt_name
+					(cert[i], j, buffer, &s, 0);
+
+				if (t == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE)
+					break;
+
+				if (t != GNUTLS_SAN_DNSNAME)
+					continue;
+
+				if (s == sizeof(buffer))
+					--s; /* The API is not clear */
+
+				buffer[s]=0;
+				(*dump_func)("Subject-Alt-Name-DNS: ", -1,
+					     dump_arg);
+				(*dump_func)(buffer, -1, dump_arg);
+				(*dump_func)("\n", -1, dump_arg);
+			}
 
 #if 0
 			(*dump_func)("Issuer:\n", -1, dump_arg);
