@@ -1,5 +1,5 @@
 /*
-** Copyright 2000-2014 Double Precision, Inc.
+** Copyright 2000-2016 Double Precision, Inc.
 ** See COPYING for distribution information.
 */
 #include	"config.h"
@@ -9,6 +9,7 @@
 #include	"libcouriertls.h"
 #include	<openssl/rand.h>
 #include	<openssl/x509.h>
+#include	<openssl/x509v3.h>
 #include	"tlscache.h"
 #include	"rfc1035/rfc1035.h"
 #include	"soxwrap/soxwrap.h"
@@ -21,6 +22,7 @@
 #include	<stdlib.h>
 #include	<ctype.h>
 #include	<netdb.h>
+#include	<idna.h>
 #if HAVE_DIRENT_H
 #include <dirent.h>
 #define NAMLEN(dirent) strlen((dirent)->d_name)
@@ -62,20 +64,19 @@ struct proto_ops {
 };
 struct proto_ops op_list[] =
 {
-#ifdef HAVE_TLSV1_2_METHOD
+#ifdef SSL_OP_NO_TLSv1
+#ifdef SSL_OP_NO_TLSv1_1
     { "TLSv1.2+",  &SSLv23_method,  SSL_OP_ALL|SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3|SSL_OP_NO_TLSv1|SSL_OP_NO_TLSv1_1 },
-    { "TLSv1.2",   &TLSv1_2_method, SSL_OP_ALL },
+    { "TLSv1.2",   &SSLv23_method,  SSL_OP_ALL|SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3|SSL_OP_NO_TLSv1|SSL_OP_NO_TLSv1_1 },
 #endif
-#ifdef HAVE_TLSV1_1_METHOD
+#endif
+
+#ifdef SSL_OP_NO_TLSv1
     { "TLSv1.1+",  &SSLv23_method,  SSL_OP_ALL|SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3|SSL_OP_NO_TLSv1 },
-    { "TLSv1.1",   &TLSv1_1_method, SSL_OP_ALL },
+    { "TLSv1.1",   &SSLv23_method,  SSL_OP_ALL|SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3|SSL_OP_NO_TLSv1 },
 #endif
-    { "TLSv1+",    &SSLv23_method,  SSL_OP_ALL|SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3 },
-    { "TLSv1",     &TLSv1_method,   SSL_OP_ALL },
-    { "TLS1",      &TLSv1_method,   SSL_OP_ALL },
-    { "SSL3+",     &SSLv23_method,  SSL_OP_ALL|SSL_OP_NO_SSLv2 },
-    { "SSL3",      &SSLv3_method,   SSL_OP_ALL },
-    { "SSL23",     &SSLv23_method,  SSL_OP_ALL },
+    { "TLSv1",     &SSLv23_method,  SSL_OP_ALL|SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3 },
+    { "TLS1",      &SSLv23_method,  SSL_OP_ALL|SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3 },
     { "",          &SSLv23_method,  SSL_OP_ALL|SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3 },
     { NULL,        &SSLv23_method,  SSL_OP_ALL|SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3 },
 };
@@ -141,38 +142,141 @@ static int ssl_verify_callback(int goodcert, X509_STORE_CTX *x509)
 	return (1);
 }
 
+static int hostmatch_utf8(const char *a, const char *b)
+{
+	while (*a || *b)
+	{
+		char ca=*a;
+		char cb=*b;
+
+		if (ca >= 'A' && ca <= 'Z')
+			ca += 'a'-'A';
+		if (cb >= 'A' && cb <= 'Z')
+			cb += 'a'-'A';
+		if (ca != cb)
+			return 0;
+
+		++a;
+		++b;
+	}
+	return 1;
+}
+
+static int hostmatch(const struct tls_info *info, const char *domain)
+{
+	const char *p=domain;
+	const char *verify_domain=info->peer_verify_domain;
+	char *idn_domain1;
+	char *idn_domain2;
+	int rc;
+
+	if (*p == '*')
+	{
+		++p;
+
+		if (*p != '.')
+			return 0;
+
+		while (*verify_domain)
+		{
+			if (*verify_domain++ == '.')
+				break;
+		}
+	}
+
+	if (idna_to_unicode_8z8z(info->peer_verify_domain, &idn_domain1, 0)
+	    != IDNA_SUCCESS)
+		idn_domain1=0;
+
+	if (idna_to_unicode_8z8z(p, &idn_domain2, 0)
+	    != IDNA_SUCCESS)
+		idn_domain2=0;
+
+	rc=hostmatch_utf8(idn_domain1 ? idn_domain1:info->peer_verify_domain,
+			  idn_domain2 ? idn_domain2:p);
+
+	if (idn_domain1)
+		free(idn_domain1);
+	if (idn_domain2)
+		free(idn_domain2);
+	return rc;
+}
+
+static int verifypeer1(const struct tls_info *info, X509 *x,
+		       STACK_OF(GENERAL_NAME) *subject_alt_names);
+
 static int verifypeer(const struct tls_info *info, SSL *ssl)
 {
 	X509 *x=NULL;
-	X509_NAME *subj=NULL;
-	int nentries, j;
-	char domain[256];
-	char *p;
-	char errmsg[1000];
+	int rc;
+	STACK_OF(GENERAL_NAME) *subject_alt_names;
 
 	if (!info->peer_verify_domain)
 		return (1);
 
-	if (info->isserver)
-	{
-		x=SSL_get_peer_certificate(ssl);
+	if (SSL_get_verify_result(ssl) != X509_V_OK)
+		return (1);
 
-		if (x)
-			subj=X509_get_subject_name(x);
-	}
-	else
-	{
-		STACK_OF(X509) *peer_cert_chain=SSL_get_peer_cert_chain(ssl);
+	x=SSL_get_peer_certificate(ssl);
 
-		if (peer_cert_chain && peer_cert_chain->stack.num > 0)
+	if (!x)
+		return (0);
+
+	subject_alt_names=
+		X509_get_ext_d2i(x, NID_subject_alt_name, NULL, NULL);
+
+	rc=verifypeer1(info, x, subject_alt_names);
+
+	if (subject_alt_names)
+		GENERAL_NAMES_free(subject_alt_names);
+
+	X509_free(x);
+
+	return rc;
+}
+
+static int verifypeer1(const struct tls_info *info, X509 *x,
+		       STACK_OF(GENERAL_NAME) *subject_alt_names)
+{
+	X509_NAME *subj=NULL;
+	int nentries, j;
+	char domain[256];
+	char errmsg[1000];
+
+	if(subject_alt_names)
+	{
+		int n=sk_GENERAL_NAME_num(subject_alt_names);
+		int i;
+
+		for (i=0; i<n; i++)
 		{
-			X509 *xx=(X509 *)peer_cert_chain->stack.data[0];
+			const GENERAL_NAME *gn=
+				sk_GENERAL_NAME_value(subject_alt_names, i);
+			const char *str;
+			int l;
 
-			if (xx)
-				subj=X509_get_subject_name(xx);
+			if (gn->type != GEN_DNS)
+				continue;
+
+#ifdef HAVE_OPENSSL110
+			str = (const char *)ASN1_STRING_get0_data(gn->d.ia5);
+#else
+			str = (const char *)ASN1_STRING_data(gn->d.ia5);
+#endif
+			l=ASN1_STRING_length(gn->d.ia5);
+
+			if (l >= sizeof(domain)-1)
+				l=sizeof(domain)-1;
+
+			memcpy(domain, str, l);
+			domain[l]=0;
+
+			if (hostmatch(info, domain))
+				return 1;
 		}
 	}
 
+	subj=X509_get_subject_name(x);
 
 	nentries=0;
 	if (subj)
@@ -187,7 +291,7 @@ static int verifypeer(const struct tls_info *info, SSL *ssl)
 		ASN1_STRING *d;
 
 		int dlen;
-		unsigned char *ddata;
+		const unsigned char *ddata;
 
 		e=X509_NAME_get_entry(subj, j);
 		if (!e)
@@ -202,8 +306,11 @@ static int verifypeer(const struct tls_info *info, SSL *ssl)
 		obj_name=OBJ_nid2sn(OBJ_obj2nid(o));
 
 		dlen=ASN1_STRING_length(d);
+#ifdef HAVE_OPENSSL110
+		ddata=ASN1_STRING_get0_data(d);
+#else
 		ddata=ASN1_STRING_data(d);
-
+#endif
 		if (strcasecmp(obj_name, "CN") == 0)
 		{
 			if (dlen >= sizeof(domain)-1)
@@ -214,23 +321,8 @@ static int verifypeer(const struct tls_info *info, SSL *ssl)
 		}
 	}
 
-	if (x)
-		X509_free(x);
-	p=domain;
-
-	if (*p == '*')
-	{
-		int	pl, l;
-
-		pl=strlen(++p);
-		l=strlen(info->peer_verify_domain);
-
-		if (*p == '.' && pl <= l &&
-		    strcasecmp(info->peer_verify_domain+l-pl, p) == 0)
-			return (1);
-	}
-	else if (strcasecmp(info->peer_verify_domain, p) == 0)
-		return (1);
+	if (domain[0] && hostmatch(info, domain))
+		return 1;
 
 	strcpy(errmsg, "couriertls: Mismatched SSL certificate: CN=");
 	strcat(errmsg, domain);
@@ -306,17 +398,107 @@ static void load_dh_params(SSL_CTX *ctx, const char *filename,
 			DH_free(dh);
 		}
 		else
-			sslerror(info, filename, -1);
+		{
+			/*
+			** If the certificate file does not have DH parameters,
+			** swallow the error.
+			*/
+
+			int err=ERR_peek_last_error();
+
+			if (ERR_GET_LIB(err) == ERR_LIB_PEM
+			    && ERR_GET_REASON(err) == PEM_R_NO_START_LINE)
+			{
+				ERR_clear_error();
+			}
+			else
+			{
+				sslerror(info, filename, -1);
+			}
+		}
 		BIO_free(bio);
 	}
 	else
 		sslerror(info, filename, -1);
 }
 
+static int check_readable_file(const char *filename)
+{
+	return (access(filename, R_OK) == 0) ? 1 : 0;
+}
+
+#ifdef HAVE_OPENSSL_SNI
+static char *get_servername_concated_readable_file(const char *filename,
+				   const char *servername,
+				   struct tls_info *info)
+{
+	char *filename_buffer;
+	char *p;
+
+	if (!filename || !servername) return NULL;
+	if (!*filename || !*servername) return NULL;
+
+	filename_buffer=malloc(strlen(filename)+strlen(servername)+2);
+	if (!filename_buffer)
+	{
+		nonsslerror(info, "malloc");
+		exit(1);
+	}
+
+	strcat(strcpy(filename_buffer, filename), ".");
+
+	p=filename_buffer + strlen(filename_buffer);
+
+	while ((*p=*servername) != 0)
+	{
+		if (*p == '/')
+			*p='.'; /* Script kiddie check */
+		++p;
+		++servername;
+	}
+	if (check_readable_file(filename_buffer)) {
+		return filename_buffer;
+	}
+
+	free(filename_buffer);
+	return NULL;
+}
+#endif
+
+static char *get_ip_concated_readable_file(SSL_CTX *ctx, const char *filename, const char *ip)
+{
+	char *test_file;
+	const struct tls_info *info=SSL_CTX_get_app_data(ctx);
+
+	if (!filename || !ip) return NULL;
+	if (!*filename || !*ip) return NULL;
+
+	test_file= malloc(strlen(filename)+strlen(ip)+2);
+	if (!test_file)
+	{
+		nonsslerror(info, "malloc");
+		exit(1);
+	}
+
+	strcpy(test_file, filename);
+	strcat(test_file, ".");
+	strcat(test_file, ip);
+
+	if (check_readable_file(test_file)) {
+		return test_file;
+	}
+
+	free(test_file);
+	return NULL;
+}
+
 static int read_certfile(SSL_CTX *ctx, const char *filename,
+			      const char *private_key_filename,
 			      int *cert_file_flags)
 {
 	const struct tls_info *info=SSL_CTX_get_app_data(ctx);
+	const char *privatekey_file = (private_key_filename == NULL)
+				    ? filename : private_key_filename;
 
 	if(!SSL_CTX_use_certificate_chain_file(ctx, filename))
 	{
@@ -326,7 +508,7 @@ static int read_certfile(SSL_CTX *ctx, const char *filename,
 
 	load_dh_params(ctx, filename, cert_file_flags);
 
-	if(!SSL_CTX_use_PrivateKey_file(ctx, filename, SSL_FILETYPE_PEM))
+	if(!SSL_CTX_use_PrivateKey_file(ctx, privatekey_file, SSL_FILETYPE_PEM))
 	{
 		sslerror(info, filename, -1);
 		return (0);
@@ -334,36 +516,39 @@ static int read_certfile(SSL_CTX *ctx, const char *filename,
 	return (1);
 }
 
-static int process_certfile(SSL_CTX *ctx, const char *certfile, const char *ip,
+static int process_certfile(SSL_CTX *ctx, const char *certfile,
+			    const char *private_key_file,
+			    const char *ip,
 			    int (*func)(SSL_CTX *, const char *,
+					const char *,
 					int *),
 			    int *cert_file_flags)
 {
 	if (ip && *ip)
 	{
 		char *test_file;
+		char *test_private_key_file;
 
 		if (strncmp(ip, "::ffff:", 7) == 0 && strchr(ip, '.'))
-			return (process_certfile(ctx, certfile, ip+7, func, cert_file_flags));
+			return (process_certfile(ctx, certfile, private_key_file, ip+7, func, cert_file_flags));
 
-		test_file= malloc(strlen(certfile)+strlen(ip)+2);
-
-		strcpy(test_file, certfile);
-		strcat(test_file, ".");
-		strcat(test_file, ip);
-
-		if (access(test_file, R_OK) == 0)
+		test_file= get_ip_concated_readable_file(ctx, certfile, ip);
+		test_private_key_file = get_ip_concated_readable_file(ctx, private_key_file, ip);
+		if (test_file != NULL)
 		{
-			int rc= (*func)(ctx, test_file,
+			int rc= (*func)(ctx, test_file, test_private_key_file,
 					cert_file_flags);
 
 			free(test_file);
+			if (test_private_key_file) free(test_private_key_file);
+
 			return rc;
 		}
-		free(test_file);
+		if (test_private_key_file) free(test_private_key_file);
 	}
 
-	return (*func)(ctx, certfile, cert_file_flags);
+	private_key_file = check_readable_file(private_key_file) ? private_key_file : NULL;
+	return (*func)(ctx, certfile, private_key_file, cert_file_flags);
 }
 
 static int client_cert_cb(ssl_handle ssl, X509 **x509, EVP_PKEY **pkey)
@@ -412,16 +597,15 @@ static int client_cert_cb(ssl_handle ssl, X509 **x509, EVP_PKEY **pkey)
 			continue;
 		}
 
-		for (i=0; client_cas && i<client_cas->stack.num; i++)
+		for (i=0; client_cas && i<sk_X509_NAME_num(client_cas); i++)
 		{
-			X509_NAME *cert=(X509_NAME *)client_cas->stack.data[i];
+			X509_NAME *cert=sk_X509_NAME_value(client_cas, i);
 
-			if (X509_NAME_cmp(cert,
-					  x->cert_info->issuer) == 0)
+			if (X509_NAME_cmp(cert, X509_get_issuer_name(x)) == 0)
 				break;
 		}
 
-		if (!client_cas || i >= client_cas->stack.num)
+		if (!client_cas || i >= sk_X509_NAME_num(client_cas))
 		{
 			BIO_free(certbio);
 			continue;
@@ -474,7 +658,63 @@ static int client_cert_cb(ssl_handle ssl, X509 **x509, EVP_PKEY **pkey)
 	return rc;
 }
 
+static SSL_CTX *tls_create_int(int isserver, const struct tls_info *info,
+			       int internal);
+
+static int server_cert_cb(ssl_handle ssl, int *ad, void *arg)
+{
+#ifdef HAVE_OPENSSL_SNI
+	struct tls_info *info=(struct tls_info *)SSL_get_app_data(ssl);
+	const char *servername=SSL_get_servername(ssl,
+						  TLSEXT_NAMETYPE_host_name);
+	const char *certfile=safe_getenv(info, "TLS_CERTFILE");
+	const char *private_keyfile=safe_getenv(info, "TLS_PRIVATE_KEYFILE");
+	int cert_file_flags=0;
+	char *cert_file_buffer;
+	char *private_keyfile_buffer;
+
+	if (!servername || !certfile)
+		return SSL_TLSEXT_ERR_OK;
+
+	cert_file_buffer = get_servername_concated_readable_file(certfile, servername, info);
+	private_keyfile_buffer = get_servername_concated_readable_file(private_keyfile, servername, info);
+	if (cert_file_buffer != NULL)
+	{
+		SSL_CTX *orig_ctx=SSL_get_SSL_CTX(ssl);
+		SSL_CTX *temp_ctx=tls_create_int(1, info, 1);
+		int rc;
+
+		if (!temp_ctx)
+		{
+			(*info->tls_err_msg)("Cannot load certificate file",
+					     info->app_data);
+			exit(1);
+		}
+		SSL_set_SSL_CTX(ssl, temp_ctx);
+		rc=read_certfile(orig_ctx, cert_file_buffer, private_keyfile_buffer, &cert_file_flags);
+		SSL_set_SSL_CTX(ssl, orig_ctx);
+		tls_destroy(temp_ctx);
+		if (!rc)
+		{
+			(*info->tls_err_msg)("Cannot load certificate file",
+					     info->app_data);
+			exit(1);
+		}
+	}
+	if (cert_file_buffer) free(cert_file_buffer);
+	if (private_keyfile_buffer) free(private_keyfile_buffer);
+
+#endif
+	return SSL_TLSEXT_ERR_OK;
+}
+
 SSL_CTX *tls_create(int isserver, const struct tls_info *info)
+{
+	return tls_create_int(isserver, info, 0);
+}
+
+SSL_CTX *tls_create_int(int isserver, const struct tls_info *info,
+			int internal)
 {
 	SSL_CTX *ctx;
 	const char *protocol=safe_getenv(info, "TLS_PROTOCOL");
@@ -482,6 +722,7 @@ SSL_CTX *tls_create(int isserver, const struct tls_info *info)
 	int session_timeout=atoi(safe_getenv(info, "TLS_TIMEOUT"));
 	const char *dhparamsfile=safe_getenv(info, "TLS_DHPARAMS");
 	const char *certfile=safe_getenv(info, "TLS_CERTFILE");
+	const char *private_keyfile=safe_getenv(info, "TLS_PRIVATE_KEYFILE");
 	const char *s;
 	struct stat stat_buf;
 	const char *peer_cert_dir=NULL;
@@ -499,6 +740,9 @@ SSL_CTX *tls_create(int isserver, const struct tls_info *info)
 	if (!*certfile)
 		certfile=NULL;
 
+	if (!*private_keyfile)
+		private_keyfile=NULL;
+
 	if (!*dhparamsfile)
 		dhparamsfile=NULL;
 
@@ -523,8 +767,10 @@ SSL_CTX *tls_create(int isserver, const struct tls_info *info)
 		if (first)
 		{
 			first=0;
+#if OPENSSL_API_COMPAT < 0x10100000L
 			SSL_load_error_strings();
 			SSLeay_add_ssl_algorithms();
+#endif
 
 			while (RAND_status() != 1)
 			{
@@ -589,8 +835,15 @@ SSL_CTX *tls_create(int isserver, const struct tls_info *info)
 #endif
 #endif
 	info_copy->tlscache=NULL;
-	init_session_cache(info_copy, ctx);
 
+	if (internal)
+	{
+		SSL_CTX_set_session_cache_mode(ctx, SSL_SESS_CACHE_OFF);
+	}
+	else
+	{
+		init_session_cache(info_copy, ctx);
+	}
 
 	s = safe_getenv(info, "TCPLOCALIP");
 
@@ -599,7 +852,7 @@ SSL_CTX *tls_create(int isserver, const struct tls_info *info)
 	if (dhparamsfile)
 		load_dh_params(ctx, dhparamsfile, &cert_file_flags);
 
-	if (certfile && !process_certfile(ctx, certfile, s,
+	if (certfile && !process_certfile(ctx, certfile, private_keyfile, s,
 					  read_certfile,
 					  &cert_file_flags))
 	{
@@ -607,8 +860,6 @@ SSL_CTX *tls_create(int isserver, const struct tls_info *info)
 		return (NULL);
 	}
 
-	SSL_CTX_set_session_cache_mode(ctx, SSL_SESS_CACHE_BOTH);
-
 	n=atoi(safe_getenv(info, "TLS_INTCACHESIZE"));
 
 	if (n > 0)
@@ -669,13 +920,11 @@ SSL_CTX *tls_create(int isserver, const struct tls_info *info)
 					      "/"), de->d_name);
 
 				fp=fopen(q, "r");
-				if (!fp)
-				{
-					nonsslerror(info, q);
-					exit(1);
-				}
 				free(q);
 
+				if (!fp)
+					continue;
+
 				while ((x=PEM_read_X509(fp, NULL, NULL, NULL)))
 				{
 					SSL_CTX_add_client_CA(ctx,x);
@@ -687,10 +936,18 @@ SSL_CTX *tls_create(int isserver, const struct tls_info *info)
 				closedir(dirp);
                 }
 	}
+
 	SSL_CTX_set_verify(ctx, get_peer_verify_level(info),
 			   ssl_verify_callback);
-	if (!isserver)
+
+	if (isserver)
+	{
+		SSL_CTX_set_tlsext_servername_callback(ctx, server_cert_cb);
+	}
+	else
+	{
 		SSL_CTX_set_client_cert_cb(ctx, client_cert_cb);
+	}
 	return (ctx);
 }
 
@@ -712,8 +969,14 @@ void tls_destroy(SSL_CTX *ctx)
 
 static int cache_add(SSL *ssl, SSL_SESSION *sess);
 
+#ifdef HAVE_OPENSSL110
+static SSL_SESSION *cache_get(SSL *ssl, const unsigned char *id, int id_len,
+			      int *copyflag);
+#else
 static SSL_SESSION *cache_get(SSL *ssl, unsigned char *id, int id_len,
 			      int *copyflag);
+#endif
+
 static void cache_del(SSL_CTX *ctx, SSL_SESSION *ssl);
 
 static void init_session_cache(struct tls_info *info, SSL_CTX *ctx)
@@ -755,10 +1018,12 @@ static int cache_add(SSL *ssl, SSL_SESSION *sess)
 	unsigned char *ucp;
 	time_t timeout= (time_t)SSL_SESSION_get_time(sess)
 		+ SSL_SESSION_get_timeout(sess);
-	void *session_id=(void *)sess->session_id;
-	size_t session_id_len=sess->session_id_length;
+	unsigned int session_id_len;
+	void *session_id;
 	size_t sess_len=i2d_SSL_SESSION(sess, NULL);
 
+	session_id=(void *)SSL_SESSION_get_id(sess, &session_id_len);
+
 	if (sizeof(timeout) + sizeof(session_id_len) + session_id_len +
 	    sess_len > sizeof(buffer))
 	{
@@ -788,7 +1053,7 @@ static int cache_add(SSL *ssl, SSL_SESSION *sess)
 }
 
 struct walk_info {
-	unsigned char *id;
+	const unsigned char *id;
 	int id_len;
 	int *copyflag;
 	SSL_SESSION *ret;
@@ -798,8 +1063,13 @@ struct walk_info {
 static int get_func(void *rec, size_t recsize,
 		    int *doupdate, void *arg);
 
+#ifdef HAVE_OPENSSL110
+static SSL_SESSION *cache_get(SSL *ssl, const unsigned char *id, int id_len,
+			      int *copyflag)
+#else
 static SSL_SESSION *cache_get(SSL *ssl, unsigned char *id, int id_len,
 			      int *copyflag)
+#endif
 {
 	const struct tls_info *info=SSL_get_app_data(ssl);
 	struct walk_info wi;
@@ -864,11 +1134,14 @@ static void cache_del(SSL_CTX *ctx, SSL_SESSION *sess)
 {
 	const struct tls_info *info=SSL_CTX_get_app_data(ctx);
 	struct walk_info wi;
+	unsigned int session_id_len;
+	void *session_id;
 
 	wi.now=0;
 
-	wi.id=(unsigned char *)sess->session_id;
-	wi.id_len=sess->session_id_length;
+	session_id=(void *)SSL_SESSION_get_id(sess, &session_id_len);
+	wi.id=(unsigned char *)session_id;
+	wi.id_len=session_id_len;
 	if (tls_cache_walk(info->tlscache, del_func, &wi) < 0)
 		perror("ALERT: tls_cache_walk: ");
 }
@@ -995,6 +1268,13 @@ SSL *tls_connect(SSL_CTX *ctx, int fd)
 	{
 		SSL_set_connect_state(ssl);
 
+#ifdef HAVE_OPENSSL_SNI
+		if (info->peer_verify_domain)
+		{
+			SSL_set_tlsext_host_name(ssl, info->peer_verify_domain);
+		}
+#endif
+
 		if ((rc=SSL_connect(ssl)) > 0)
 		{
 			if (!verifypeer(info, ssl))
@@ -1032,7 +1312,10 @@ void tls_disconnect(SSL *ssl, int fd)
 	fcntl(fd, F_SETFL, 0);
 	SSL_set_shutdown(ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
 	SSL_free(ssl);
+#ifdef HAVE_OPENSSL110
+#else
 	ERR_remove_state(0);
+#endif
 }
 
 /* --------------------------------------- */
@@ -1281,6 +1564,8 @@ static void dump_x509(X509 *x509,
 	X509_NAME *subj=X509_get_subject_name(x509);
 	int nentries, j;
 	time_t timestamp;
+	STACK_OF(GENERAL_NAME) *subject_alt_names;
+
 	static const char gcc_shutup[]="%Y-%m-%d %H:%M:%S";
 
 	if (!subj)
@@ -1297,7 +1582,7 @@ static void dump_x509(X509 *x509,
 		ASN1_STRING *d;
 
 		int dlen;
-		unsigned char *ddata;
+		const unsigned char *ddata;
 
 		e=X509_NAME_get_entry(subj, j);
 		if (!e)
@@ -1312,7 +1597,11 @@ static void dump_x509(X509 *x509,
 		obj_name=OBJ_nid2sn(OBJ_obj2nid(o));
 
 		dlen=ASN1_STRING_length(d);
+#ifdef HAVE_OPENSSL110
+		ddata=ASN1_STRING_get0_data(d);
+#else
 		ddata=ASN1_STRING_data(d);
+#endif
 
 		(*dump_func)("   ", -1, dump_arg);
 		(*dump_func)(obj_name, -1, dump_arg);
@@ -1323,6 +1612,39 @@ static void dump_x509(X509 *x509,
 	}
 	(*dump_func)("\n", 1, dump_arg);
 
+	subject_alt_names=
+		X509_get_ext_d2i(x509, NID_subject_alt_name, NULL, NULL);
+
+	if (subject_alt_names)
+	{
+		int n=sk_GENERAL_NAME_num(subject_alt_names);
+		int i;
+
+		for (i=0; i<n; i++)
+		{
+			const GENERAL_NAME *gn=
+				sk_GENERAL_NAME_value(subject_alt_names, i);
+			const char *domain;
+			int l;
+
+			if (gn->type != GEN_DNS)
+				continue;
+
+#ifdef HAVE_OPENSSL110
+			domain = (const char *)ASN1_STRING_get0_data(gn->d.ia5);
+#else
+			domain = (const char *)ASN1_STRING_data(gn->d.ia5);
+#endif
+			l=ASN1_STRING_length(gn->d.ia5);
+
+			(*dump_func)("Subject-Alt-Name-DNS: ", -1, dump_arg);
+			(*dump_func)(domain, l, dump_arg);
+			(*dump_func)("\n", -1, dump_arg);
+		}
+
+		GENERAL_NAMES_free(subject_alt_names);
+	}
+
 	timestamp=asn1toTime(X509_get_notBefore(x509));
 
 	if (timestamp)
@@ -1375,8 +1697,9 @@ void tls_dump_connection_info(ssl_handle ssl,
 			}
 		}
 
-		for (i=0; peer_cert_chain && i<peer_cert_chain->stack.num; i++)
-			dump_x509((X509 *)peer_cert_chain->stack.data[i],
+		for (i=0; peer_cert_chain && i<sk_X509_num(peer_cert_chain);
+		     i++)
+			dump_x509((X509 *)sk_X509_value(peer_cert_chain, i),
 				  dump_func, dump_arg);
 	}
 
@@ -1554,7 +1877,8 @@ char *tls_cert_name(const char *buf, size_t buf_size)
 
 	if (x)
 	{
-		p=X509_NAME_oneline(x->cert_info->subject, NULL, 0);
+		X509_get_subject_name(x);
+		p=X509_NAME_oneline(X509_get_subject_name(x), NULL, 0);
 		X509_free(x);
 	}
 	ERR_clear_error();
